/*

 gzilla

 Copyright 1997 Raph Levien <raph@acm.org>

 This code is free for commercial and non-commercial use,
 modification, and redistribution, as long as the source code release,
 startup screen, or product packaging includes this copyright notice.

 */

/* The gif decoder for gzilla. It is responsible for decoding GIF data
   in a bytesink and transferring it to an imgsink.

   Adapted by Raph Levien from giftopnm.c as found in the
   netpbm-1mar1994 release. The copyright notice for giftopnm.c
   follows: */

/* +-------------------------------------------------------------------+ */
/* | Copyright 1990, 1991, 1993, David Koblas.  (koblas@netcom.com)    | */
/* |   Permission to use, copy, modify, and distribute this software   | */
/* |   and its documentation for any purpose and without fee is hereby | */
/* |   granted, provided that the above copyright notice appear in all | */
/* |   copies and that both that copyright notice and this permission  | */
/* |   notice appear in supporting documentation.  This software is    | */
/* |   provided "as is" without express or implied warranty.           | */
/* +-------------------------------------------------------------------+ */

/* Todos:

 + Make sure to handle error cases gracefully (including aborting the
   imgsink, if necessary).

 */

#include <string.h> /* for memcpy and memmove */

#include <gtk/gtk.h>
#include "gzillabytesink.h"
#include "gzillaimgsink.h"
#include "gzillagif.h"

static void gzilla_gif_class_init (GzillaGifClass *klass);
static void gzilla_gif_init       (GzillaGif      *gif);
static void gzilla_gif_destroy    (GtkObject      *object);
static void gzilla_gif_write (GzillaByteSink *bytesink,
			      char *bytes,
			      gint num);

static GzillaByteSinkClass *parent_class = NULL;



#define INTERLACE              0x40
#define LOCALCOLORMAP  0x80

#define LM_to_uint(a,b)                        (((b)<<8)|(a))

#define GZILLA_GIF_CLEAN 0
#define GZILLA_GIF_DIRTY 1
#define GZILLA_GIF_SIZECHANGE 2

guint
gzilla_gif_get_type ()
{
  static guint gif_type = 0;

  if (!gif_type)
    {
      GtkTypeInfo gif_info =
      {
	"GzillaGif",
	sizeof (GzillaGif),
	sizeof (GzillaGifClass),
	(GtkClassInitFunc) gzilla_gif_class_init,
	(GtkObjectInitFunc) gzilla_gif_init,
	(GtkArgFunc) NULL,
      };

      gif_type = gtk_type_unique (gzilla_bytesink_get_type (), &gif_info);
    }

  return gif_type;
}

void
gzilla_gif_class_init (GzillaGifClass *class)
{
  GtkObjectClass      *object_class;
  GzillaByteSinkClass *bytesink_class;

  parent_class = gtk_type_class (gzilla_bytesink_get_type ());

  object_class = (GtkObjectClass*) class;
  bytesink_class = (GzillaByteSinkClass*) class;

  object_class->destroy = gzilla_gif_destroy;

  bytesink_class->write = gzilla_gif_write;
}

void gzilla_gif_init (GzillaGif *gif) {
  gif->state = 0;
  gif->size_inbuf = 0;
  gif->size_inbuf_max = 1024;
  gif->inbuf = g_new (char, gif->size_inbuf_max);
  gif->linebuf = NULL;
  gif->fresh = FALSE;
  gif->transparent = -1;
}

static gint process_bytes (GzillaGif *gif, char *buf, gint bufsize);

static void lwz_init (GzillaGif *gif);
static void get_code_init (GzillaGif *gif);
static int
lwz_read_byte (GzillaGif *gif, char **pbuf, gint *pbufsize);
static int
get_code (GzillaGif *gif, char **pbuf, gint *pbufsize);
static int
get_data_block (char **pbuf, gint *pbufsize, char *buf);

GzillaByteSink*
gzilla_gif_new (GzillaImgSink *imgsink)
{
  GzillaGif *gif;

  gif = gtk_type_new (gzilla_gif_get_type ());
  gif->bytesink.widget = NULL;

  gif->imgsink = imgsink;
  return GZILLA_BYTESINK (gif);
}

static void gzilla_gif_write (GzillaByteSink *bytesink,
			      char *bytes,
			      gint num) {
  GzillaGif *gif;
  char *buf;
  gint bufsize, bytes_consumed;

  gif = (GzillaGif *)bytesink;

#ifdef VERBOSE
  g_print ("gzilla_gif_write: %d bytes\n", num);
#endif
  /* Concatenate with the partial input, if any. */
  if (gif->size_inbuf) {
    bufsize = gif->size_inbuf + num;
    while (gif->size_inbuf_max < bufsize)
      gif->size_inbuf_max <<= 1;
    gif->inbuf = g_realloc (gif->inbuf, gif->size_inbuf_max);
    buf = gif->inbuf;
    memcpy (buf + gif->size_inbuf, bytes, num);
  } else {
    buf = bytes;
    bufsize = num;
  }

  /* Process the bytes in the input buffer. */
  do {
    if (bufsize > 0) {
      bytes_consumed = process_bytes (gif, buf, bufsize);
      buf += bytes_consumed;
      bufsize -= bytes_consumed;
    } else {
      bytes_consumed = 0;
    }
  } while (bytes_consumed > 0);

  /* Put unprocessed input in inbuf, if any. */
  gif->size_inbuf = bufsize;
  if (gif->size_inbuf > 0) {
    /* todo: check that inbuf is big enough */
    memmove (gif->inbuf, buf, gif->size_inbuf);
  }

}

static void gzilla_gif_destroy (GtkObject *object) {
  GzillaGif *gif;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GZILLA_IS_GIF (object));

  gif = GZILLA_GIF (object);
  if (gif->linebuf != NULL)
    g_free (gif->linebuf);
  g_free (gif->inbuf);

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static gboolean read_ok (char **pbuf, gint *pbufsize, char *obuf, gint size) {
  if (size > *pbufsize)
    return FALSE;
  memcpy (obuf, *pbuf, size);
  *pbuf += size;
  *pbufsize -= size;
  return TRUE;
}

/* Return TRUE when the extension is over */
static gboolean do_extension (GzillaGif *gif, char **pbuf, gint *pbufsize) {
  gint size;
  unsigned char buf[260];

  size = get_data_block (pbuf, pbufsize, buf);
  if (size == -1)
    return FALSE;
  else if (size == 0)
    return TRUE;
  if (gif->label == 0xf9) {
    /* Graphic Control Extension */
    gif->disposal = (buf[0] >> 2) & 0x7;
    gif->inputFlag = (buf[0] >> 1) & 0x1;
    gif->delayTime = LM_to_uint (buf[1], buf[2]);
    if (buf[0] & 0x1)
      gif->transparent = buf[3];
  }
  return FALSE;
}

/* Process some bytes from the input gif stream. It's a state machine. */
gint process_bytes (GzillaGif *gif, char *ibuf, gint bufsize) {
  gchar *tmp_buf = ibuf;
  gint tmp_bufsize = bufsize;
  unsigned char buf[16];
  char c;
  int byte;
  gint initial_state;
  unsigned char *cmap;

  initial_state = gif->state;

  switch (gif->state) {

  case 0:
    /* at beginning of file - read magic number */
    if (!read_ok (&tmp_buf, &tmp_bufsize, buf, 6))
      break;
    if (strncmp ((char *)buf, "GIF", 3) != 0) {
      gif->state = 999;
      break;
    }
    if (strncmp ((char *)buf + 3, "87a", 3) != 0 &&
	strncmp ((char *)buf + 3, "89a", 3) != 0) {
      gif->state = 999;
      break;
    }
    gif->state = 1;
    break;

  case 1:
    /* screen descriptor */
    if (!read_ok (&tmp_buf, &tmp_bufsize, buf, 7))
      break;
    gif->Width = LM_to_uint(buf[0],buf[1]);
    gif->Height = LM_to_uint(buf[2],buf[3]);
    gif->BitPixel = 2<<(buf[4]&0x07);
    gif->ColorResolution = (((buf[4]&0x70)>>3)+1);
    gif->Background      = buf[5];
    gif->AspectRatio     = buf[6];

    gif->linebuf = g_new (char, gif->Width);
    gzilla_imgsink_set_parms (gif->imgsink, gif->Width, gif->Height,
			      GZILLA_IMG_TYPE_INDEXED);

    if (buf[4] & LOCALCOLORMAP) {
      gif->state = 2;
    } else {
      gif->state = 3;
    }
    break;

  case 2:
    /* read colormap */
    if (!read_ok (&tmp_buf, &tmp_bufsize,
		  gif->ColorMap,
		  gif->BitPixel * 3))
      break;
    gif->state = 3;
    break;

  case 3:
    /* next character is one of:
       ';' = GIF terminator
       '!' = extension
       ',' = start character
       */
    if (!read_ok (&tmp_buf, &tmp_bufsize, &c, 1))
      break;
#ifdef VERBOSE
    g_print ("c = %c\n", c);
#endif
    if (c == ',') {
      gif->state = 4;
    } else if (c == '!') {
      gif->state = 7;
    } else {
      gif->state = 999;
    }
    break;

  case 4:
    /* right after the start character */
    if (!read_ok (&tmp_buf, &tmp_bufsize, buf, 9))
      break;
    /* we should probably just check that the local stuff is consistent
       with the stuff at the header. For now, we punt... */
    gif->interlace = (buf[8] & INTERLACE) != 0;
    gif->pass = 0;
    gif->state = 5;
    break;

  case 5:
    /* corresponds to ReadImage */
    if (!read_ok (&tmp_buf, &tmp_bufsize, &gif->input_code_size, 1))
      break;
    gif->x = 0;
    gif->y = 0;
    lwz_init (gif);
    if (gif->imgsink)
      gzilla_imgsink_set_cmap (gif->imgsink, gif->ColorMap, 256,
			       gif->transparent);
    gif->state = 6;
    break;

  case 6:
    /* get an image byte */
    cmap = gif->ColorMap;
    do {
      byte = lwz_read_byte (gif, &tmp_buf, &tmp_bufsize);
      if (byte < 0)
	break;
      if (gif->imgsink) {
	gif->linebuf[gif->x] = byte;
      } else {
	gif->linebuf[gif->x * 3] = cmap[byte * 3];
	gif->linebuf[gif->x * 3 + 1] = cmap[byte * 3 + 1];
	gif->linebuf[gif->x * 3 + 2] = cmap[byte * 3 + 2];
      }
      gif->x++;
      if (gif->x == gif->Width) {
	gzilla_imgsink_write (gif->imgsink, gif->linebuf, 0, gif->y, 1);

	gif->x = 0;
	if (gif->interlace) {
	  switch (gif->pass) {
	  case 0:
	  case 1:
	    gif->y += 8;
	    break;
	  case 2:
	    gif->y += 4;
	    break;
	  case 3:
	    gif->y += 2;
	    break;
	  }
	  if (gif->y >= gif->Height) {
	    gif->pass++;
	    switch (gif->pass) {
	    case 1:
	      gif->y = 4;
	      break;
	    case 2:
	      gif->y = 2;
	      break;
	    case 3:
	      gif->y = 1;
	      break;
	    }
	  }

	} else {
	  gif->y++;
	}
      }
    } while (c >= 0 && gif->y < gif->Height);
    if (gif->y >= gif->Height) {
      gzilla_imgsink_close (gif->imgsink);
      gif->state = 999;
    }
    break;

  case 7:
    /* get the extension type */
    if (!read_ok (&tmp_buf, &tmp_bufsize, &gif->label, 1))
      break;
    gif->state = 8;
    break;

  case 8:
    /* corresponds to DoExtension */
    if (do_extension (gif, &tmp_buf, &tmp_bufsize)) {
      gif->state = 3;
    }
    break;

  default:
    /* error - just consume all input */
    tmp_bufsize = 0;
  }
#if 0
  printf ("process_bytes: initial state %d, final state %d, %d bytes consumed\n",
	  initial_state, gif->state, bufsize - tmp_bufsize);
#endif
  return bufsize - tmp_bufsize;
}

static void
lwz_init (GzillaGif *gif) {
  int i;

  gif->set_code_size = gif->input_code_size;
  gif->code_size = gif->set_code_size+1;
  gif->clear_code = 1 << gif->set_code_size;
  gif->end_code = gif->clear_code + 1;
  gif->max_code_size = 2*gif->clear_code;
  gif->max_code = gif->clear_code+2;

  get_code_init (gif);

  gif->fresh = TRUE;

  for (i = 0; i < gif->clear_code; ++i) {
    gif->table[0][i] = 0;
    gif->table[1][i] = i;
  }
  for (; i < (1<<MAX_LWZ_BITS); ++i)
    gif->table[0][i] = gif->table[1][0] = 0;

  gif->sp = gif->stack;
}

static int
lwz_read_byte (GzillaGif *gif, char **pbuf, gint *pbufsize) {
  int    code, incode;
  int    i;

  if (gif->fresh) {
    do {
      gif->firstcode = gif->oldcode =
	get_code(gif, pbuf, pbufsize);
    } while (gif->firstcode == gif->clear_code);
    if (gif->firstcode >= 0)
      gif->fresh = FALSE;
    return gif->firstcode;
  }

  if (gif->sp > gif->stack)
    return *--gif->sp;

  while ((code = get_code(gif, pbuf, pbufsize)) >= 0) {
    if (gif->firstcode < 0) {
      gif->firstcode = gif->oldcode = code;
      return gif->firstcode;
    } else if (code == gif->clear_code) {
      for (i = 0; i < gif->clear_code; ++i) {
	gif->table[0][i] = 0;
	gif->table[1][i] = i;
      }
      for (; i < (1<<MAX_LWZ_BITS); ++i)
	gif->table[0][i] = gif->table[1][i] = 0;
      gif->code_size = gif->set_code_size+1;
      gif->max_code_size = 2*gif->clear_code;
      gif->max_code = gif->clear_code+2;
      gif->sp = gif->stack;
      gif->firstcode = gif->oldcode =
	get_code(gif, pbuf, pbufsize);
      return gif->firstcode;
    } else if (code == gif->end_code) {
      int             count;
      unsigned char   buf[260];

      if (gif->done)
	return -2;

      while ((count = get_data_block(pbuf, pbufsize, buf)) > 0)
	;

#if 0
      if (count != 0)
	pm_message("missing EOD in data stream (common occurence)");
      /* of course, the other possibility is that the EOD
	 spans a block boundary */
#endif
      return -2;
    }

    incode = code;

    if (code >= gif->max_code) {
      *gif->sp++ = gif->firstcode;
      code = gif->oldcode;
    }

    while (code >= gif->clear_code) {
      *gif->sp++ = gif->table[1][code];
#if 0
      if (code == gif->table[0][code])
	pm_error("circular table entry BIG ERROR");
#endif
      code = gif->table[0][code];
    }

    *gif->sp++ = gif->firstcode = gif->table[1][code];

    if ((code = gif->max_code) <(1<<MAX_LWZ_BITS)) {
      gif->table[0][code] = gif->oldcode;
      gif->table[1][code] = gif->firstcode;
      ++gif->max_code;
      if ((gif->max_code >= gif->max_code_size) &&
	  (gif->max_code_size < (1<<MAX_LWZ_BITS))) {
	gif->max_code_size *= 2;
	++gif->code_size;
      }
    }

    gif->oldcode = incode;

    if (gif->sp > gif->stack)
      return *--gif->sp;
  }
  return code;
}

static void get_code_init (GzillaGif *gif) {
  gif->curbit = 0;
  gif->lastbit = 0;
  gif->done = FALSE;
}

static int
get_code (GzillaGif *gif, char **pbuf, gint *pbufsize) {
       int           i, j, ret;
       int           count;
       int           code_size;
       unsigned char *buf;

       code_size = gif->code_size;
       buf = gif->buf;

       /* Why is this inequality nonstrict? */
       if ( (gif->curbit + code_size) >= gif->lastbit) {
               if (gif->done) {
#if 0
		 /* should probably handle this error condition */
                       if (curbit >= lastbit)
                               pm_error("ran off the end of my bits" );
#endif
                       return -1;
               }

	       count = get_data_block(pbuf, pbufsize, &buf[2]);
	       if (count < 0)
		 return -1;
               else if (count == 0)
		 gif->done = TRUE;

               gif->last_byte = 2 + count;
               gif->curbit = (gif->curbit - gif->lastbit) + 16;
               gif->lastbit = (2+count)*8 ;
       }

       ret = 0;
       for (i = gif->curbit, j = 0; j < code_size; ++i, ++j)
	 ret |= ((buf[i >> 3] & (1 << (i & 7))) != 0) << j;

       gif->curbit += code_size;

       if ( (gif->curbit + code_size) >= gif->lastbit) {
	 buf[0] = buf[gif->last_byte-2];
	 buf[1] = buf[gif->last_byte-1];
       }

       return ret;
}

/* Return number of bytes in block or -1 if no data present. */
static int
get_data_block (char **pbuf, gint *pbufsize, char *buf) {
  int size;

  if (*pbufsize > 0) {
    size = *(unsigned char *)*pbuf;
    if (*pbufsize >= 1 + size) {
      memcpy (buf, *pbuf + 1, size);
      *pbuf += 1 + size;
      *pbufsize -= 1 + size;
      return size;
    } else {
      return -1;
    }
  } else {
    return -1;
  }
}
